import gzip
import sys
from .nullcontext import nullcontext

def choose_open_function(filename, force_gzip=None):
    '''
    If `force_gzip` is True or False, use corresponding open function.
    If `force_gzip` is None, it's guessed based on filename extension
    '''
    if (force_gzip is True) or ((force_gzip is None) and filename.endswith('.gz')):
        return gzip.open
    elif (force_gzip is False) or ((force_gzip is None) and not filename.endswith('.gz')):
        return open
    else:
        raise ValueError("`force_gzip` should be one of True/False/None")

def open_for_write(filename, force_gzip=None, mode='wt', **kwargs):
    if filename and (filename != '-'):
        open_func = choose_open_function(filename=filename, force_gzip=force_gzip)
        return open_func(filename, mode, **kwargs)
    else:
        return nullcontext(sys.stdout)

def open_for_read(filename, force_gzip=None, mode='rt', **kwargs):
    if filename and (filename != '-'):
        open_func = choose_open_function(filename=filename, force_gzip=force_gzip)
        return open_func(filename, mode, **kwargs)
    else:
        return nullcontext(sys.stdin)



# GFF parser adapted from https://techoverflow.net/2013/11/30/a-simple-gff3-parser-in-python/

from collections import namedtuple
import gzip
import json
from .gzip_utils import open_for_read

_gff_info_fields = ["contig", "source", "type", "start", "stop", "score", "strand", "phase", "attributes"]
class GTFRecord(namedtuple("GTFRecord", _gff_info_fields)):
    @property
    def length(self):
        return self.stop - self.start

    def __repr__(self):
        row = [self.contig, self.source, self.type, self.start, self.stop, self.score, self.strand, self.phase, self.encoded_attributes()]
        row = [elem if elem else '.'  for elem in row]
        return '\t'.join(map(str, row))

    def contain_position(self, pos):
        return self.start <= pos < self.stop

    def in_upstream_of(self, pos):
        if self.strand == '+':
            return self.stop <= pos
        elif self.strand == '-':
            return pos < self.start

    def in_downstream_of(self, pos):
        if self.strand == '+':
            return pos < self.start
        elif self.strand == '-':
            return self.stop <= pos

    def attributes_filtered(self, relevant_attributes):
        '''
        Use to drop unnecessary attributes and make a program more memory-efficient:
          record.attributes_filtered({'attr_1', 'attr_2', ...})
        Returns a new record with modified attributes
        '''
        attributes_filtered = {k: v  for (k,v) in self.attributes.items()  if k in relevant_attributes}
        return GTFRecord(self.contig, self.source, self.type, self.start, self.stop, self.score, self.strand, self.phase, attributes_filtered)

    def attributes_renamed(self, attr_mapping):
        '''
        Use to rename incorrectly named attributes:
          record.attributes_renamed({'wrong_attr_name': 'correct_attr_name', ...})
        Returns a new record with modified attributes; unspecified attributes are left unmodified
        '''
        attributes_modified = {}
        for (orig_attr_name, attr_value) in self.attributes.items():
            if orig_attr_name in attr_mapping:
                attr_name = attr_mapping[orig_attr_name]
            else:
                attr_name = orig_attr_name
            attributes_modified[attr_name] = attr_value
        return GTFRecord(self.contig, self.source, self.type, self.start, self.stop, self.score, self.strand, self.phase, attributes_modified)

    def encoded_attributes(self):
        return self.encode_gtf_attributes(self.attributes)

    @classmethod
    def encode_gtf_attributes(cls, attributes):
        if not attributes:
            return '.'
        attr_strings = [f'{k} {json.dumps(v)};' for k,v in attributes.items()]
        return ' '.join(attr_strings)

    @classmethod
    def each_in_file(cls, filename, multivalue_keys=None):
        """
        A minimalistic GTF format parser.
        Yields objects that contain info about a single GTF feature.

        Supports transparent gzip decompression.
        """
        with open_for_read(filename, encoding='utf-8') as infile:
            for line in infile:
                if line.startswith("#"): continue
                parts = line.strip().split("\t")
                # If this fails, the file format is not standard-compatible
                assert len(parts) == len(_gff_info_fields)
                assert parts[0] != '.'  # contig
                assert parts[2] != '.'  # type
                assert parts[3] != '.'  # start
                assert parts[4] != '.'  # stop
                assert parts[6] in {'+', '-'}

                # Normalize data
                normalized_info = {
                    "contig": parts[0],
                    "source": None if parts[1] == "." else parts[1],
                    "type":   parts[2],
                    "start":  int(parts[3]) - 1, # 0-based, included
                    "stop":   int(parts[4]),     # 0-based, excluded
                    "score": None if parts[5] == "." else float(parts[5]),
                    "strand": parts[6],
                    "phase": None if parts[7] == "." else parts[7],
                    "attributes": cls.parse_gtf_attributes(parts[8], multivalue_keys=multivalue_keys),
                }
                yield GTFRecord(**normalized_info)

    @classmethod
    def parse_gtf_attributes(cls, attribute_string, multivalue_keys=None):
        """Parse the GTF attribute column and return a dict"""
        if attribute_string == ".":
            return {}
        ret = {}
        for attribute in attribute_string.strip().rstrip(";").split(";"):
            key, value = attribute.strip().split(" ", maxsplit=1)

            if value[0] == '"':
                val = value[1:-1]
            else:
                val = int(value)

            # Some records has several attributes with the same key
            # (e.g. `tag tag_1; tag tag_2;`)
            # We treat values for such keys as lists
            if key not in multivalue_keys:
                if key not in ret:
                    ret[key] = val
                else:
                    raise Exception(f'Key `{key}` already in attributes.\n'
                                     'Probably you should add this attribute to a set of `multivalue_keys`')
            else:
                if key not in ret:
                    ret[key] = []
                ret[key].append(val)
        return ret

# deprecated
def parse_gtf(filename, multivalue_keys=None):
    yield from GTFRecord.each_in_file(filename, multivalue_keys=multivalue_keys)



import dataclasses
from typing import List, Any
from .dataclass_tsv_serializable import DataclassTsvSerializable

@dataclasses.dataclass(order=True, frozen=True)
class Interval(DataclassTsvSerializable):
    '''bed-coordinates [a, b)'''
    chrom: str
    start: int
    stop: int
    rest: List[Any] = dataclasses.field(default_factory=list, metadata={'skip_conversion': True})

    def __post_init__(self):
        if not isinstance(self.chrom, str):
            raise TypeError(f'Chromosome should be a string but was `{self.chrom}`')
        if not isinstance(self.start, int):
            raise TypeError(f'Interval start should be integer but was `{self.start}`')
        if not isinstance(self.stop, int):
            raise TypeError(f'Interval stop should be integer but was `{self.stop}`')
        if self.start >= self.stop:
            raise ValueError(f'Interval start={self.start} should be less than stop={self.stop}')

    @classmethod
    def from_string(cls, line, **kwargs):
        row = line.rstrip('\n').split('\t')
        chrom, start, stop, *rest = row
        attrs = {'chrom': chrom, 'start': int(start), 'stop': int(stop), 'rest': rest}
        return cls(**attrs, **kwargs)

    def tsv_string(self):
        fields = [self.chrom, self.start, self.stop, *self.rest]
        return '\t'.join(map(str, fields))

    @property
    def length(self):
        return self.stop - self.start



from collections import defaultdict
import itertools
import pybedtools
from .gtf_parser import GTFRecord
from .utils import take_the_only
from .fasta_reader import fasta_from_file
from .dto.coding_transcript_info import CodingTranscriptInfo

class Annotation:
    def __init__(self):
        self.gene_by_id = {}
        self.transcript_by_id = {}
        self.geneId_by_transcript = {}
        self.transcripts_by_gene = defaultdict(list)
        self.parts_by_transcript = defaultdict(list)

    @classmethod
    def load(cls, filename, relevant_attributes=None, coding_only=False, attr_mapping=None, multivalue_keys=None):
        annotation = cls()

        if multivalue_keys is None:
            # multivalued keys for Gencode data model:
            # https://www.gencodegenes.org/pages/data_format.html
            multivalue_keys = {'tag', 'ont', 'ccdsid'}

        records = GTFRecord.each_in_file(filename, multivalue_keys=multivalue_keys)
        if attr_mapping:
            records = map(lambda rec: rec.attributes_renamed(attr_mapping), records)

        if relevant_attributes is not None:
            # that's the minimum list of attributes for a library to properly work
            necessary_attributes = {'gene_id', 'transcript_id', 'gene_type', 'transcript_type'}
            relevant_attributes = relevant_attributes.union(necessary_attributes)
            records = map(lambda rec: rec.attributes_filtered(relevant_attributes), records)

        if coding_only:
            records = filter(cls.is_coding, records)

        for rec in records:
            annotation.push_record(rec)
        return annotation

    def push_record(self, rec):
        gene_id = rec.attributes['gene_id']
        if rec.type == 'gene':
            self.gene_by_id[gene_id] = rec
        else:
            transcript_id = rec.attributes['transcript_id']
            if rec.type == 'transcript':
                self.transcript_by_id[transcript_id] = rec
                self.geneId_by_transcript[transcript_id] = gene_id
                self.transcripts_by_gene[gene_id].append(rec)
            else:
                self.parts_by_transcript[transcript_id].append(rec)

    def transcript_exons(self, transcript_id):
        parts = self.parts_by_transcript[transcript_id]
        return [part  for part in parts  if part.type == 'exon']

    def transcript_cds(self, transcript_id):
        parts = self.parts_by_transcript[transcript_id]
        return [part  for part in parts  if part.type == 'CDS']

    def transcript_start_codons(self, transcript_id):
        '''
        We return a list of codons because some transcripts
        don't have an annotated start/stop codon or have
        several segments due to splicing
        '''
        parts = self.parts_by_transcript[transcript_id]
        start_codons = [part  for part in parts  if part.type == 'start_codon']
        return start_codons

    def transcript_stop_codons(self, transcript_id):
        '''
        We return a list of codons because some transcripts
        don't have an annotated start/stop codon or have
        several segments due to splicing
        '''
        parts = self.parts_by_transcript[transcript_id]
        stop_codons = [part  for part in parts  if part.type == 'stop_codon']
        return stop_codons

    def transcript_utrs(self, transcript_id):
        parts = self.parts_by_transcript[transcript_id]
        return [part  for part in parts  if part.type == 'UTR']

    def transcript_strand(self, transcript_id):
        return self.segments_strand(self.parts_by_transcript[transcript_id])

    @classmethod
    def segments_strand(cls, segments):
        strands = { segment.strand  for segment in  segments }
        if len(strands) != 1:
            raise ValueError('Different strands')
        return strands.pop()

    @classmethod
    def segments_ordered_5_to_3(cls, segments):
        strand = cls.segments_strand(segments)
        reverse_order = (strand == '-')
        by_start_coordinate = lambda segment: segment.start
        return sorted(segments, key=by_start_coordinate, reverse=reverse_order)

    def coding_transcript_info(self, transcript_id):
        """Return coding transcript info. This method fails when called for non-coding transcript"""
        gene_id = self.geneId_by_transcript[transcript_id]
        exons = self.transcript_exons(transcript_id)
        cds_segments = self.transcript_cds(transcript_id)
        strand = self.transcript_strand(transcript_id)

        transcript_length = sum(exon.length for exon in exons)
        len_cds_in_transcript = sum(cds.length for cds in cds_segments)
        if strand == '+':
            genomic_cds_start = min([rec.start for rec in cds_segments])
            len_exons_before_cds = sum( exon.length  for exon in exons  if exon.in_upstream_of(genomic_cds_start) )
            exon_with_cds = take_the_only([exon for exon in exons  if exon.contain_position(genomic_cds_start)])
            cds_offset = genomic_cds_start - exon_with_cds.start  # offset of CDS relative to embracing exon
        elif strand == '-':
            genomic_cds_start = max([rec.stop - 1 for rec in cds_segments])
            len_exons_before_cds = sum( exon.length  for exon in exons  if exon.in_upstream_of(genomic_cds_start) )
            exon_with_cds = take_the_only([exon for exon in exons  if exon.contain_position(genomic_cds_start)])
            cds_offset = exon_with_cds.stop - 1 - genomic_cds_start  # offset of CDS relative to embracing exon

        cds_start = cds_offset + len_exons_before_cds
        cds_stop = cds_start + len_cds_in_transcript

        return CodingTranscriptInfo(gene_id, transcript_id, transcript_length, cds_start, cds_stop)

    def ordered_segments_by_type(self, transcript_id, feature_type):
        if feature_type == 'exons':
            return self.segments_ordered_5_to_3(self.transcript_exons(transcript_id))
        elif feature_type == 'cds':
            return self.segments_ordered_5_to_3(self.transcript_cds(transcript_id))
        elif feature_type == 'cds_with_stop':
            return self.segments_ordered_5_to_3(self.transcript_cds(transcript_id) + self.transcript_stop_codons(transcript_id))
        elif feature_type == 'full':
            # full, unspliced transcript
            return [self.transcript_by_id(transcript_id)]


    def segments_as_bedtool_intervals(self, segments, name='.'):
        yield from (pybedtools.Interval(s.contig, s.start, s.stop, strand=s.strand, name=name) for s in segments)

    # annotation.transcript_sequence('ENSMUST00000115529.7', './genomes/mm10.fa', feature_type='cds')
    def transcript_sequence(self, transcript_id, assembly_fasta_fn, feature_type='exons'):
        iterator = self.transcript_sequences([transcript_id], assembly_fasta_fn, feature_type)
        transcript_id, sequence = next(iterator)
        return sequence

    # feature type is one of full/exons/cds
    def transcript_sequences(self, transcript_ids, assembly_fasta_fn, feature_type='exons'):
        intervals = itertools.chain.from_iterable(
            self.segments_as_bedtool_intervals(
                self.ordered_segments_by_type(transcript_id, feature_type),
                name = transcript_id
            ) for transcript_id in transcript_ids
        )
        bed = pybedtools.BedTool(intervals)
        seq_result = bed.sequence(fi = assembly_fasta_fn, name=True, s=True)  # s - strandedness
        sequences = fasta_from_file(seq_result.seqfn)
        sequence_groups = itertools.groupby(sequences, key=lambda hdr_seq_pair: hdr_seq_pair[0])
        # we have separate sequences for each feature (such as exon) of transcript, so we should join them
        for transcript_id, hdr_seq_pairs in sequence_groups:
            if transcript_id[-3:] in ['(+)', '(-)']:
                transcript_id = transcript_id[0:-3]
            sequence = ''.join(hdr_seq_pair[1] for hdr_seq_pair in hdr_seq_pairs)
            yield (transcript_id, sequence)

    @classmethod
    def is_coding(cls, record):
        if record.attributes['gene_type'] != 'protein_coding':
            return False
        if (record.type != 'gene') and (record.attributes['transcript_type'] != 'protein_coding'):
            return False
        return True



import dataclasses
import itertools
import numpy as np
from .coverage_interval import CoverageInterval

@dataclasses.dataclass
class TranscriptCoverage:
    transcript_id: str
    coverage: np.array

    @classmethod
    def each_in_file(cls, filename, header=False, dtype=float):
        bedgraph_stream = CoverageInterval.each_in_file(filename, header=False)
        yield from cls.each_in_bedgraph(bedgraph_stream, dtype=dtype)

    @classmethod
    def each_in_bedgraph(cls, bedgraph_stream, dtype=float):
        for (transcript_id, bedgraph_intervals_iter) in itertools.groupby(bedgraph_stream, lambda interval: interval.chrom):
            bedgraph_intervals = list(bedgraph_intervals_iter)
            transcript_length = max(interval.stop for interval in bedgraph_intervals)
            profile = np.zeros(transcript_length, dtype=dtype)
            for interval in bedgraph_intervals:
                # Note! pybedtools use 0-based coordinates (when in integer representation).
                # See https://daler.github.io/pybedtools/3-brief-examples.html and https://daler.github.io/pybedtools/intervals.html#zero-based-coords
                profile[interval.start : interval.stop] = interval.coverage
            yield cls(transcript_id, profile)



from sklearn.linear_model import LinearRegression
import numpy as np
from math import log

from collections import namedtuple
WeigthedPoints = namedtuple('WeigthedPoints', ['xs', 'ys', 'weights'])

def segmentation_stops(segmentation):
    stops = [s.start for s in segmentation.segments]
    stops.append(segmentation.segments[-1].stop)
    return stops

def segmentwise_sums(segmentation, profile):
    profile_cumsum = np.hstack([0, np.cumsum(profile)])
    return np.diff(profile_cumsum[segmentation_stops(segmentation)])

    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)


def slope_by_profiles(control_profile, experiment_profile, segmentation, log_mode=False):
    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)
    return slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=log_mode)

def slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=False):
    total_coverage_control = np.sum(control_sums)
    total_coverage_experiment = np.sum(experiment_sums)

    assert len(segmentation.segments) == len(control_sums) == len(experiment_sums)
    profile_len = segmentation.segmentation_length

    center_info = WeigthedPoints([], [], [])
    weighted_center_info = WeigthedPoints([], [], [])
    every_point_info = WeigthedPoints([], [], [])

    num_segments = segmentation.num_segments
    if log_mode:
        rate_transform = log
    else:
        rate_transform = lambda x: x
    for (segment, control_segment_sum, experiment_segment_sum) in zip(segmentation.segments, control_sums, experiment_sums):
        mean_control = control_segment_sum / segment.length
        mean_experiment = experiment_segment_sum / segment.length
        normalized_mean_control = (mean_control + 1) / (total_coverage_control + num_segments)
        normalized_mean_experiment = (mean_experiment + 1) / (total_coverage_experiment + num_segments)
        detrended_profile = normalized_mean_experiment / normalized_mean_control
        value = rate_transform(detrended_profile)

        coord = (segment.start + segment.stop - 1) / 2
        rel_coord = coord / profile_len

        # mode: 'center'
        center_info.xs.append(rel_coord)
        center_info.ys.append(value)
        center_info.weights.append(1)

        # mode: 'weighted_center'
        weight = segment.stop - segment.start
        weighted_center_info.xs.append(rel_coord)
        weighted_center_info.ys.append(value)
        weighted_center_info.weights.append(weight)

        # mode: 'every_point'
        for pos in range(segment.start, segment.stop):
            rel_pos = pos / profile_len
            every_point_info.xs.append(rel_pos)
            every_point_info.ys.append(value)
            every_point_info.weights.append(1)

    return {
        'center': slope_by_points(center_info.xs, center_info.ys, center_info.weights),
        'weighted_center': slope_by_points(weighted_center_info.xs, weighted_center_info.ys, weighted_center_info.weights),
        'every_point': slope_by_points(every_point_info.xs, every_point_info.ys, every_point_info.weights),
    }

def slope_by_points(xs, ys, weights):
    xs = np.array(xs)
    model = LinearRegression()
    model.fit(xs.reshape(-1, 1), ys, sample_weight=weights)
    return model.coef_[0]

def profile_difference(control_profile, experiment_profile, segmentation):
    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)
    return profile_difference(control_sums, experiment_sums, segmentation, log_mode=log_mode)

def profile_difference_by_segment_counts(control_sums, experiment_sums, segmentation):
    assert len(control_sums) == len(experiment_sums) == len(segmentation.segments)

    control_profile_sum = np.sum(control_sums)
    experiment_profile_sum = np.sum(experiment_sums)

    control_normed = control_sums / control_profile_sum  if control_profile_sum != 0  else control_sums
    experiment_normed = experiment_sums / experiment_profile_sum  if experiment_profile_sum != 0  else experiment_sums
    difference = 0
    for (segment, control_segment_sum, experiment_segment_sum) in zip(segmentation.segments, control_normed, experiment_normed):
        difference += abs(control_segment_sum - experiment_segment_sum)
    return difference / segmentation.segmentation_length



import argparse
from .version import __version__
from .bin import get_coverage, pool_coverages, extract_cds_annotation, clip_cds, \
                 coverage_properties, join_tables, choose_best, \
                 compare_coverages, \
                 shrinkage, plot_distribution, \
                 flatten_coverage

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="papolarity", description = "Main entrypoint of papolarity")
    argparser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    subparsers = argparser.add_subparsers(metavar='subcommand')
    
    subparser_configs = [
        {'cmd': 'get_coverage', 'namespace': get_coverage, 'help': 'Generates coverage from an alignment'},
        {'cmd': 'pool_coverages', 'namespace': pool_coverages, 'help': 'Pool coverage profiles'},
        {'cmd': 'extract_cds_annotation', 'namespace': extract_cds_annotation, 'help': 'Extract CDS annotation in transcriptomic coordinates from genomic annotation'},
        {'cmd': 'clip_cds', 'namespace': clip_cds, 'help': 'Clip any bed file in transcriptomic coordinates to CDS-region'},
        {'cmd': 'coverage_properties', 'namespace': coverage_properties, 'help': 'Calculate coverage profile properties'},
        {'cmd': 'join_tables', 'namespace': join_tables, 'help': 'Join the second table to the first one'},
        {'cmd': 'choose_best', 'namespace': choose_best, 'help': 'Choose best element from each group (e.g. best transcipt for each gene)'},
        {'cmd': 'compare_coverages', 'namespace': compare_coverages, 'help': 'Coverage profile comparison'},
        {'cmd': 'plot_distribution', 'namespace': plot_distribution, 'help': 'Plot distributions of properties'},
        {'cmd': 'shrinkage', 'namespace': shrinkage, 'help': 'Make length-dependend shrinkage of slope properties'},
        {'cmd': 'flatten_coverage', 'namespace': flatten_coverage, 'help': 'Flatten coverage profiles by averaging data through given segments'},
    ]
    for subparser_config in subparser_configs:
        invocation_fn = subparser_config['namespace'].invoke
        configurator = subparser_config['namespace'].configure_argparser
        subparser = subparsers.add_parser(subparser_config['cmd'], help=subparser_config['help'])
        subparser.set_defaults(invocation_fn=invocation_fn)
        configurator(subparser)
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    args.invocation_fn(args)

if __name__ == '__main__':
    main()



def fasta_from_file(filename):
    '''Yields pairs (header, sequence). Properly handles multiline FASTA.'''
    with open(filename) as fasta_file:
        header = None
        for line in fasta_file:
            line = line.rstrip('\n')
            if line.startswith('>'):
                if header != None:
                    yield (header, ''.join(sequences))
                header = line[1:].lstrip()
                sequences = []
            else:
                sequences.append(line.strip())
        if header != None:
            yield (header, ''.join(sequences))



import sys
import argparse
import itertools
from ..gzip_utils import open_for_write
from ..tsv_reader import stream_table_column_highlighted

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(
            prog = "choose_best",
            description = "Choose best element from each group (e.g. best transcipt for each gene)",
        )
    argparser.add_argument('table', metavar='table.tsv', help='Input table in tab-separated format')
    argparser.add_argument('column', help="Take the biggest or the lowest transcript according to this column")
    argparser.add_argument('criteria', choices=['max', 'min'], help="Criteria to choose transcript: `max` takes the biggest, `min` - the lowest")
    argparser.add_argument('--group-by', default='gene_id', dest='group_by_column', help="Column to group transcripts")
    
    header_group = argparser.add_mutually_exclusive_group(required=True)
    header_group.add_argument('--header', action='store_true', dest='has_header', help="Table has header")
    header_group.add_argument('--no-header', action='store_false', dest='has_header', help="Tables doesn't have header")

    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--dtype', choices=['int', 'float'], default='float', help="Treat values as int or float (default: float)")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.dtype == 'int':
        dtype = int
    elif args.dtype == 'float':
        dtype = float
    else:
        raise ValueError('dtype should be either int or float')

    get_group = lambda row: row[0][0]
    get_value = lambda row: dtype(row[0][1])
    get_row = lambda row: row[1]

    with open_for_write(args.output_file) as output_stream:
        table_stream = stream_table_column_highlighted(args.table, [args.group_by_column, args.column], has_header=args.has_header, pop_column=False)
        if args.has_header:
            _, header = next(table_stream)
            print('\t'.join(header), file=output_stream)
        
        data = list(table_stream)
        data.sort(key=get_group)
        for group, rows in itertools.groupby(data, key=get_group):
            if args.criteria == 'max':
                row = get_row(max(rows, key=get_value))
            elif args.criteria == 'min':
                row = get_row(min(rows, key=get_value))
            else:
                raise ValueError(f'Unknown criteria `{criteria}` (only min/max allowed)')
            print('\t'.join(row), file=output_stream)



from decimal import Decimal
import itertools
import numpy as np
from . import coreutils_sort
from .dto.transcript_coverage import TranscriptCoverage
from .dto.coverage_interval import CoverageInterval

def transcript_coverages_from_alignment(alignment, sort_transcripts=False, dtype=float):
    bedgraph = make_coverage(alignment, sort_transcripts=sort_transcripts, dtype=dtype)
    coverage_intervals = coverage_intervals_from_bedgraph(bedgraph, dtype=dtype)
    yield from TranscriptCoverage.each_in_bedgraph(coverage_intervals, dtype=dtype)

def make_coverage(alignment, sort_transcripts=False, stream=True, dtype=float):
    if sort_transcripts:
        # sort by chromosome (transcript_id in case of transcriptomic alignments)
        bedgraph = alignment.genome_coverage(bga=True).coreutils_sort(key=['1,1', '2,2n'], stream=stream)
    else:
        bedgraph = alignment.genome_coverage(bga=True, stream=stream)
    return bedgraph

def coverage_intervals_from_bedgraph(bedgraph, dtype=float):
    for interval in bedgraph:
        yield CoverageInterval(*interval[0:4], dtype=dtype)



import sys
import dataclasses
from ..gzip_utils import open_for_read, open_for_write

@dataclasses.dataclass
class DataclassTsvSerializable:
    # Additional columns (goes after normal ones) which are to be printed/stored but not loaded'''
    computable_properties = [] # [('column_name', 'property_name'), ...]

    @classmethod
    def header(cls):
        fields = [field.name for field in dataclasses.fields(cls)] + [name for (name, prop) in cls.computable_properties]
        return '\t'.join(fields)

    def __str__(self):
        return self.tsv_string()

    def tsv_string(self):
        fields = [getattr(self, field.name) for field in dataclasses.fields(self)] + [getattr(self, prop) for (name, prop) in self.computable_properties]
        return '\t'.join(map(str, fields))

    @classmethod
    def from_string(cls, line, **kwargs):
        row = line.rstrip('\n').split('\t')
        attrs = {}
        for field, value in zip(dataclasses.fields(cls), row): # skips computable (and other auxiliary) properties
            if field.metadata.get('skip_conversion', False):
                attrs[field.name] = value
            else:
                attrs[field.name] = field.type(value)
        return cls(**attrs, **kwargs)

    @classmethod
    def each_in_file(cls, filename, header=True, force_gzip=None):
        with open_for_read(filename, force_gzip=force_gzip) as f:
            if header:
                f.readline() # skip header
            for line in f:
                yield cls.from_string(line)

    @classmethod
    def print_tsv(cls, collection, file=sys.stdout, header=True):
        if header:
            print(cls.header(), file=file)
        for record in collection:
            print(record.tsv_string(), file=file)

    @classmethod
    def store_tsv(cls, collection, filename, header=True, force_gzip=None):
        with open_for_write(filename, force_gzip=force_gzip) as f:
            cls.print_tsv(collection, f, header=header)



import argparse
import math
from pybedtools import BedTool
import pasio
from ..utils import align_iterators, get_constant_intervals
from ..gzip_utils import open_for_write
from ..dto.coverage_interval import CoverageInterval
from ..dto.transcript_coverage import TranscriptCoverage
from ..segmentation import Segmentation

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(
            prog = "flatten_coverage",
            description = "Flatten coverage profiles by averaging data through given segments",
        )
    argparser.add_argument('segmentation', help='Segmentation of contigs in bed format')
    argparser.add_argument('coverage', help='Coverage in bedgraph format')
    argparser.add_argument('--only-matching', action='store_true', help="Don't pool coverages of transcripts which are present not in all files")
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--rounding', choices=['no', 'round', 'ceil', 'floor'], default='none', help="Rounding of float values (default: no rounding)")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.rounding == 'round':
        rounding = round
    elif args.rounding == 'ceil':
        rounding = math.ceil
    elif args.rounding == 'floor':
        rounding = math.floor
    elif args.rounding in ['no', 'none']:
        rounding = lambda x: x
    else:
        raise ValueError('rounding should be either int or float')

    transcript_coverage_stream = TranscriptCoverage.each_in_file(args.coverage, header=False, dtype=float)

    segmentation_stream = Segmentation.each_in_file(args.segmentation, header=False)

    with open_for_write(args.output_file) as output_stream:
        aligned_transcripts = align_iterators([segmentation_stream, transcript_coverage_stream], key=[lambda segment: segment.chrom, lambda transcript_coverage: transcript_coverage.transcript_id], check_sorted=True)
        for (transcript_id, (segmentation, transcript_coverage)) in aligned_transcripts:
            if transcript_coverage is None:
                continue
            if (segmentation is None) and args.only_matching:
                continue
            if segmentation is not None:
                transcript_coverage = segmentation.stabilize_profile(transcript_coverage.coverage)
            coverage_intervals = get_constant_intervals(transcript_coverage)
            bedgraph = (CoverageInterval(transcript_id, *interval, dtype=float) for interval in coverage_intervals)
            CoverageInterval.print_tsv(bedgraph, header=False, file=output_stream)



import csv
from .gzip_utils import open_for_read

def stream_table_column_highlighted(filename, columns, has_header, pop_column=True):
    with open_for_read(filename) as file:
        if has_header:
            header = next(file)
            header = header.rstrip("\n").split("\t")
            try:
                column_indices = [int(column) for column in columns]
            except ValueError:
                column_indices = [header.index(column) for column in columns]
            column_names = tuple([header[column_idx] for column_idx in column_indices])
            if pop_column:
                for column_idx in column_indices:
                    header.pop(column_idx)
            yield (column_names, header)
        else:
            column_idx = int(column)

        for line in file:
            row = line.rstrip("\n").split("\t")
            key = tuple([row[column_idx] for column_idx in column_indices])
            if pop_column:
                for column_idx in column_indices:
                    row.pop(column_idx)
            yield (key, row)

def each_in_tsv(filename):
    with open_for_read(filename) as input_stream:
        reader = csv.DictReader(input_stream, delimiter='\t')
        for row in reader:
            yield row



import argparse
from ..gzip_utils import open_for_write
from ..annotation import Annotation
from ..dto.coding_transcript_info import CodingTranscriptInfo

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="extract_cds_annotation", description = "Extract CDS annotation in transcriptomic coordinates from genomic annotation")
    argparser.add_argument('gtf_annotation', metavar='annotation.gtf', help='Genomic annotation in GTF-format')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    annotation = Annotation.load(args.gtf_annotation,
                                relevant_attributes=set(),
                                attr_mapping={'gene_biotype': 'gene_type', 'transcript_biotype': 'transcript_type'},
                                coding_only=True)
    with open_for_write(args.output_file) as output_stream:
        print(CodingTranscriptInfo.header(), file=output_stream)
        for transcript_id in annotation.transcript_by_id:
            cds_info = annotation.coding_transcript_info(transcript_id)
            print(cds_info, file=output_stream)



import argparse
from pybedtools import BedTool
from ..coverage_profile import make_coverage, coverage_intervals_from_bedgraph
from ..dto.coverage_interval import CoverageInterval

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="get_coverage", description="Generates coverage from an alignment")
    argparser.add_argument('alignment', help='alignment in BAM format')
    argparser.add_argument('--sort', action='store_true', help="Sort resulting alignments by transcript name")
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--dtype', choices=['int', 'float'], default='int', help="Make int or float-valued coverage (default: int)")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.dtype == 'int':
        dtype = int
    elif args.dtype == 'float':
        dtype = float
    else:
        raise ValueError('dtype should be either int or float')

    alignment = BedTool(args.alignment)
    bedgraph = make_coverage(alignment, sort_transcripts=args.sort, stream=True, dtype=dtype)

    if args.output_file:
        bedgraph.saveas(args.output_file)
    else:
        intervals = coverage_intervals_from_bedgraph(bedgraph, dtype=dtype)
        CoverageInterval.print_tsv(intervals, header=False)



import contextlib

if hasattr(contextlib, 'nullcontext'):
    nullcontext = contextlib.nullcontext
else:
    class nullcontext:
        def __init__(self, enter_result=None):
            self.enter_result = enter_result
        def __enter__(self):
            return self.enter_result
        def __exit__(self, *excinfo):
            pass



import argparse
from ..utils import common_subsequence
from ..gzip_utils import open_for_write
from ..dto.transcript_coverage import TranscriptCoverage
from ..segmentation import Segmentation
from ..dto.coverage_comparison_stats import CoverageComparisonStats

def compare_coverage_streams(segmentations, control_coverages, experiment_coverages):
    segmentation_and_coverages = [
        segmentations,
        control_coverages,
        experiment_coverages,
    ]

    key_extractors = [
        lambda segment: segment.chrom,
        lambda transcript_coverage: transcript_coverage.transcript_id,
        lambda transcript_coverage: transcript_coverage.transcript_id,
    ]

    transcript_stream = common_subsequence(segmentation_and_coverages, key=key_extractors, check_sorted=True)
    for (transcript_id, (segmentation, control_coverage, experiment_coverage)) in transcript_stream:
        yield CoverageComparisonStats.make_from_profiles(transcript_id, control_coverage.coverage, experiment_coverage.coverage, segmentation)

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(
            prog = "compare_coverages",
            description = "Coverage profile comparison",
        )
    argparser.add_argument('segmentation', metavar='segmentation.bed', help='Segmentation')
    argparser.add_argument('coverage_control', metavar='control.bedgraph', help='Coverage for control data')
    argparser.add_argument('coverage_experiment', metavar='experiment.bedgraph', help='Coverage for experiment data')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    segmentation_stream = Segmentation.each_in_file(args.segmentation, header=False)

    control_coverages = TranscriptCoverage.each_in_file(args.coverage_control, header=False, dtype=int)
    experiment_coverages = TranscriptCoverage.each_in_file(args.coverage_experiment, header=False, dtype=int)

    transcript_comparison_infos = compare_coverage_streams(segmentation_stream, control_coverages, experiment_coverages)
    with open_for_write(args.output_file) as output_stream:
        CoverageComparisonStats.print_tsv(transcript_comparison_infos, header=True, file=output_stream)



import numpy as np

def flatten(xs):
    result = []
    for x in xs:
        if isinstance(x, list):
            result.extend(flatten(x))
        else:
            result.append(x)
    return result

def take_the_only(arr):
    if len(arr) > 1:
        raise Exception('Several elements when the only one is expected')
    if len(arr) == 0:
        raise Exception('No elements when one is expected')
    return arr[0]

def common_subsequence(iterators, key=lambda x: x, check_sorted=False):
    sentinel = object()
    for (key, aligned_objects) in align_iterators(iterators, key=key, object_missing=sentinel, check_sorted=check_sorted):
        if sentinel not in aligned_objects:
            yield (key, aligned_objects)

_sentinel_key = object()
def align_iterators(iterators, key=lambda x: x, object_missing=None, check_sorted=False):
    '''
    Take a list of iterables yielding objects whose `key` values increase
    strictly monotonically. At each iteration yield a pair consisting of
    a key and a tuple of objects from these interators which have the same key.
    If one of iterators doesn't contain corresponding value or yields,
    then `object_missing` (which is `None` by default) is substituted 
    at that place.
    `key` can be either a callable object (same key for each iterator)
    or a list of callable objects (one key per iterator)
    '''
    iterators = [iter(iterator) for iterator in iterators]
    num_iters = len(iterators)
    exhausted = [False] * num_iters
    objects = [None] * num_iters
    if callable(key):
        key = [key for _ in iterators]
    else:
        assert len(iterators) == len(key)
    key_values = [_sentinel_key] * num_iters
    for idx in range(num_iters):
        objects[idx], key_values[idx], exhausted[idx] = _next_unless_exhausted(iterators[idx], key[idx], exhausted[idx], object_missing=object_missing)
    while not all(exhausted):
        min_key = min(k for k in key_values if k is not _sentinel_key)
        matching_idxs = {idx  for (idx, k) in enumerate(key_values)  if k == min_key}
        aligned_objects = [(objects[idx] if idx in matching_idxs else object_missing)  for idx in range(num_iters)]
        yield (min_key, aligned_objects)
        for idx in matching_idxs:
            old_key_value = key_values[idx]
            objects[idx], key_values[idx], exhausted[idx] = _next_unless_exhausted(iterators[idx], key[idx], exhausted[idx])
            if check_sorted and (key_values[idx] is not _sentinel_key) and (key_values[idx] <= old_key_value):
                raise ValueError(f"Iterator at index `{idx}` is not sorted: `{key_values[idx]}` follows `{old_key_value}`")

def _next_unless_exhausted(iterator, key, exhausted, object_missing=None):
    '''
    Utility function to take values from an iterator and transform them to keys
    until iterator is exhausted and report iterator's status
    '''
    if not exhausted:
        try:
            obj = next(iterator)
            k = key(obj)
            return (obj, k, False)
        except StopIteration:
            return (object_missing, _sentinel_key, True)
    else:
        return (object_missing, _sentinel_key, True)

def pool_profiles(profiles):
    return np.sum(profiles, axis=0)

def get_constant_intervals(profile):
    if len(profile) == 0:
        return []
    last_value_indices = np.nonzero(np.diff(profile))[0]
    start_indices_inclusive = np.concatenate(([0], last_value_indices + 1))
    end_indices_exclusive = np.concatenate( (last_value_indices + 1, [len(profile)]) )
    for (start, end) in zip(start_indices_inclusive, end_indices_exclusive):
        yield (start, end, profile[start])
